#include <AdaptiveSamplingTracer.h>
#include <common.h>
#include <random>

extern std::vector<Pixel> pixels;
extern unsigned int numSamplesMax;
extern ConcurrentQueue<PixelBlock> taskQueue_;
extern ImagePlane imagePlane;
ConcurrentQueue<PixelBlock> taskQueue2_;

void AdaptiveSamplingTracer::BuildQueue()
{
    // build the pixel block queue
    const int blockSize = 8;
    for (int x = 0; x < camera_.imgWidth; x += blockSize)
    {
        for (int y = 0; y < camera_.imgHeight; y += blockSize)
        {
            PixelBlock block;
            block.fromX = x;
            block.toX = std::min(x + blockSize, camera_.imgWidth);
            block.fromY = y;
            block.toY = std::min(y + blockSize, camera_.imgHeight);
            taskQueue_.Push(block);
            taskQueue2_.Push(block);
        }
    }
}

void AdaptiveSamplingTracer::operator()()
{
    FirstPass();
    SecondPass();
}

// trace indirect illumination, only 1 ray
void AdaptiveSamplingTracer::SecondPass()
{
    int i = 0;
    while (!taskQueue2_.IsEmpty())
    {
        PixelBlock block;
        taskQueue2_.PopIfNonEmpty(block);
        for (int y = block.fromY; y != block.toY; ++y)
        {
            for (int x = block.fromX; x != block.toX; ++x)
            {
                int pixId = y * camera_.imgWidth + x;
                Pixel & pixel = pixels[pixId];
                Color col(0, 0, 0);
                for (size_t j = 0; j < pixel.subpixels.size(); ++j)
                {
                    auto & subpix = pixel.subpixels[j];
                    if (j % 4 == 0)
                    {
                        col = DoRayTrace2nd(camera_, pixel, subpix);
                    }
                    subpix.color += col;
                }
            }
        }
        ++i;
        if (i % 10 == 0)
            std::cout << taskQueue2_.Size() << "_ ";
    }
}

void AdaptiveSamplingTracer::FirstPass()
{
    int i = 0;
    while (!taskQueue_.IsEmpty())
    {
        PixelBlock block;
        taskQueue_.PopIfNonEmpty(block);
        for (int y = block.fromY; y != block.toY; ++y)
        {
            for (int x = block.fromX; x != block.toX; ++x)
            {
                int pixId = y * camera_.imgWidth + x;
                Pixel & pixel = pixels[pixId];
                // do ray tracing for each subpixel
                for (size_t i = 0; i < pixel.subpixels.size(); ++i)
                {
                    DoRayTrace(camera_, pixel, pixel.subpixels[i]);
                }

                // check the variance of the average of the subpixel colors
                Color mean = pixel.GetMeanColor();
                while (pixel.subpixels.size() < numSamplesMax)
                {
                    pixel.AddSubpixel();
                    DoRayTrace(camera_, pixel, pixel.subpixels.back());
                    float numSubPix = static_cast<float>(pixel.subpixels.size());
                    Color newmean = (mean * (numSubPix - 1) + pixel.subpixels.back().color) / numSubPix;
                    Color diff = (newmean - mean);
                    mean = newmean;
                    float diffTotal = std::max(std::max(fabs(diff.r), fabs(diff.g)), fabs(diff.b));
                    if (diffTotal < 0.05f)
                    {
                        break;
                    }
                }
                pixel.color = mean;
            }
        }
        ++i;
        if (i % 10 == 0)
            std::cout << taskQueue_.Size() << " ";
    }
}