#include <AdaptiveSubdivisionTracer.h>
#include <algorithm>

float DiffColor(const Color & c1, const Color & c2)
{
    Color c = c1 - c2;
    return c.r * c.r + c.g * c.g + c.b * c.b;
}

void AdaptiveSubdivisionTracer::ProcessBlock(SubPixelBlock * b)
{
    if ((b->p[1].x - b->p[0].x) % 2 == 0 && (b->p[0].y - b->p[3].y) % 2 == 0) // can be subdivided
    {
        // find the new points
        Point2i p[5]; // top, right, down, left, center
        p[0].x = p[4].x = p[2].x = (b->p[1].x + b->p[0].x) / 2;
        p[0].y = b->p[0].y;
        p[1].x = b->p[1].x;
        p[1].y = p[4].y = p[3].y = (b->p[0].y + b->p[3].y) / 2;
        p[2].y = b->p[2].y;
        p[3].x = b->p[3].x;

        // add the sub-blocks
        b->subblocks.push_back(SubPixelBlock(b->p[0], p[0], p[4], p[3]));
        b->subblocks.push_back(SubPixelBlock(p[0], b->p[1], p[1], p[4]));
        b->subblocks.push_back(SubPixelBlock(p[4], p[1], b->p[2], p[2]));
        b->subblocks.push_back(SubPixelBlock(p[3], p[4], p[2], b->p[3]));        

        // add the new points to the list of points to trace
        pointsToTrace.insert(p[0]);
        pointsToTrace.insert(p[1]);
        pointsToTrace.insert(p[2]);
        pointsToTrace.insert(p[3]);
        pointsToTrace.insert(p[4]);

        // add the 4 new sub-blocks into the list of new blocks
        newBlocks.push_back(&b->subblocks[0]);
        newBlocks.push_back(&b->subblocks[1]);
        newBlocks.push_back(&b->subblocks[2]);
        newBlocks.push_back(&b->subblocks[3]);
    }
}

void AdaptiveSubdivisionTracer::ProcessBlocks()
{
    using std::max;
    
    concurrency::parallel_for(size_t(0), oldBlocks.size(), [this](size_t i)
    {
        SubPixelBlock * block = oldBlocks[i];
        // check the color of the four corners
        bool subdivide = false;
        for (std::size_t i = 0; i < 4; ++i)
        {
            for (std::size_t j = i + 1; j < 4; ++j)
            {
                Color diff = samples[block->p[i]] - samples[block->p[j]];
                diff.r = fabs(diff.r); diff.g = fabs(diff.g); diff.b = fabs(diff.b);
                float diffMax = max(diff.r, max(diff.g, diff.b));
                if (diffMax > 0.1f)
                {
                    subdivide = true;
                    break;
                }
            }
        }
        if (!subdivide) return;

        // process each block
        ProcessBlock(block);                        
    });        
}

void AdaptiveSubdivisionTracer::TraceSamples()
{
    // trace all samples
}

void AdaptiveSubdivisionTracer::ShadeBlocks()
{
    // shade all blocks in a hierarchical manner
}

void AdaptiveSubdivisionTracer::Trace()
{
    while (!pointsToTrace.empty())
    {
        TraceSamples();
        ProcessBlocks();
        pointsToTrace.clear();
        oldBlocks.swap(newBlocks);
        newBlocks.clear();
    }
}

void AdaptiveSubdivisionTracer::GenerateInitialBlocks()
{
    // generate all initial blocks and points to trace
}