#pragma once
#ifndef ADAPTIVE_SUBDIVISION_TRACER_H_
#define ADAPTIVE_SUBDIVISION_TRACER_H_

#pragma warning(push, 1)
#include <scene.h>
#pragma warning(pop)
#include <ppl.h>
#include <concurrent_unordered_map.h>
#include <concurrent_unordered_set.h>
#include <concurrent_vector.h>
#include <vector>
#include <RayTracer.h>

struct Point2i
{
    size_t x;
    size_t y;
    bool operator==(const Point2i & other) const
    {
        return x == other.x && y == other.y;
    }
};

struct Sample
{
    size_t x;
    size_t y;
    Color color;
};

namespace std
{
    template <>
    struct hash < Point2i >
    {
        size_t operator()(Point2i const & p) const
        {
            return (51 + std::hash<size_t>()(p.x)) * 51 + std::hash<size_t>()(p.y);
        }
    };
}

struct SubPixelBlock
{
    SubPixelBlock() { }
    SubPixelBlock(const Point2i & p0, const Point2i & p1, const Point2i & p2, const Point2i & p3)
    {
        p[0] = p0;
        p[1] = p1;
        p[2] = p2;
        p[3] = p3;
    }
    Point2i p[4]; // 4 corners
    std::vector<SubPixelBlock> subblocks; // this can either be empty or of size 4
    Color color;
};

class AdaptiveSubdivisionTracer : public RayTracer
{
private:
    std::vector<SubPixelBlock> blocks;
    concurrency::concurrent_unordered_map<Point2i, Color> samples;
    concurrency::concurrent_unordered_set<Point2i> pointsToTrace;
    concurrency::concurrent_vector<SubPixelBlock *> oldBlocks;
    concurrency::concurrent_vector<SubPixelBlock *> newBlocks;

    void Trace();
    void GenerateInitialBlocks(); // generate initial blocks and points
    void ProcessBlocks(); // process all blocks in parallel
    void ProcessBlock(SubPixelBlock * b);
    void TraceSamples(); // trace all samples
    void ShadeBlocks(); // shade all blocks
};

#endif