#include <Filtering.h>
#include <algorithm>
#include <ppl.h>
#include <common.h>

float MitchellNetravali(float x, float B, float C)
{
    static float A1 = (12 - 9 * B - 6 * C);
    static float A2 = (-18 + 12 * B + 6 * C);
    static float A3 = (6 - 2 * B);
    static float B1 = (-B - 6 * C);
    static float B2 = (6 * B + 30 * C);
    static float B3 = (-12 * B - 48 * C);
    static float B4 = (8 * B + 24 * C);
    x = std::abs(x);
    float xSquared = x * x, xCubed = xSquared * x;
    if (x < 1)
    {
        return 1.0f / 6.0f * (A1 * xCubed + A2 * xSquared + A3);
    }
    else if (x < 2)
    {
        return 1.0f / 6.0f * (B1 * xCubed + B2 * xSquared + B3 * x + B4);
    }
    return 0.0f;
}

float Gaussian(float d)
{
    const float sigma = 0.5;
    float f = 2 * sigma * sigma;
    return 1 / (PI * f) * exp(-(d * d) / f);
}

void FilterPixel(int x, int y, int width, int height, std::vector<Pixel> & image)
{
    int pixelId = y * width + x;
    float weight = 0;
    Color & color = image[pixelId].color;
    color = Color().Black();
    for (int dy = -2; dy <= 2; ++dy)
    {
        if (y + dy < 0 || y + dy >= height) continue;
        for (int dx = -2; dx <= 2; ++dx)
        {
            if (x + dx < 0 || x + dx >= width) continue;
            int neighborId = (y + dy) * width + (x + dx);
            const std::vector<SubPixel> & subpixes = image.at(neighborId).subpixels;
            for (auto & subpix : subpixes)
            {
                float distx = -0.5f + dx + subpix.rx; // distance in x
                float weightx = Gaussian(distx);//MitchellNetravali(distx);
                float disty = -0.5f + dy + subpix.ry; // distance in y
                float weighty = Gaussian(disty);// MitchellNetravali(disty);
                weight += weightx * weighty;
                color += weightx * weighty * subpix.color;
            }
        }
    }
    color /= weight;
}

void FilterImage(int width, int height, std::vector<Pixel> & image)
{
    concurrency::parallel_for(int(0), height, [&](int y)
    {
        for (int x = 0; x < width; ++x)
        {
            FilterPixel(x, y, width, height, image);
        }
    });
}