#pragma once
#ifndef PHOTON_MAPPING_H_
#define PHOTON_MAPPING_H_

#include <memory>
#include <vector>
#pragma warning (push, 1)
#include <scene.h>
#pragma warning (pop)
#include <G3D/PointHashGrid.h>

struct Photon
{
public:
    Photon() { }

    Photon(const Ray & photonRay, const Color & photonColor, const HitInfo & photonHitInfo)
    {
        dir = -photonRay.dir.GetNormalized();
        pos = photonHitInfo.p;
        col = photonColor;
    }

    Point3 pos; // hit point on surface
    Point3 dir; // negative incoming direction
    Color col; // color of the photon

    bool operator==(const Photon & other) const
    {
        return (pos == other.pos);
    }

    static void getPosition(const Photon & p, G3D::Vector3 & pos)
    {
        pos.x = p.pos.x;
        pos.y = p.pos.y;
        pos.z = p.pos.z;
    }
};

class PhotonMapping
{
public:
    PhotonMapping() { }
    void GeneratePhotonMap();
    void InitPhotonMaps();
    void FilterPhotonMap(const std::unique_ptr < PhotonMap> & inputMap, std::unique_ptr<PhotonMap> & outputMap, float size) const;
    void SavePhotonMap(const std::unique_ptr<PhotonMap> & map, const std::string & name) const;
    void LoadPhotonMap(std::unique_ptr<PhotonMap> & map, const std::string & name) const;
    void LoadLightCache();
private:
    static const unsigned int numPhotonsEmitted_ = 1000000;
    static const unsigned int numPhotonBounces = 3;
    static const bool storeFirstHit = true;
    typedef G3D::PointHashGrid<Photon, Photon> PhotonMap;

    std::vector<float> accumLightPower_;

    void AccumulateLightPower();
    const Light * SelectLight() const;
    void TracePhoton(Ray & photonRay, Color & photonColor);
    float radiusHint_;
    std::vector<Point3> lightCache_;
public:
    std::unique_ptr<PhotonMap> globalMap_;
    std::unique_ptr<PhotonMap> causticMap_;
    std::unique_ptr<PhotonMap> filteredMap_;
};

#endif