#include <Pixel.h>

extern unsigned int numSamplesMin;
extern unsigned int numSamplesMax;

Pixel::Pixel() : color(Color().Black())
{
    subpixels.resize(numSamplesMin);
    for (int i = 0; i < subpixels.size(); ++i)
    {
        subpixels[i].rx = Halton(i, basex);
        subpixels[i].ry = Halton(i, basey);
    }
}

void Pixel::AddSubpixel()
{
    subpixels.push_back(SubPixel());
    int i = static_cast<int>(subpixels.size() - 1u);
    subpixels[i].rx = Halton(i, basex);
    subpixels[i].ry = Halton(i, basey);
}

void Pixel::AddSubpixelMax()
{
    int oldSize = static_cast<int>(subpixels.size());
    subpixels.resize(numSamplesMax);
    for (int i = oldSize; i < subpixels.size(); ++i)
    {
        subpixels[i].rx = Halton(i, basex);
        subpixels[i].ry = Halton(i, basey);
    }
}

Color Pixel::GetMeanColor() const
{
    Color mean = Color().Black();
    for (auto & subpix : subpixels)
    {
        mean += subpix.color;
    }
    return mean / static_cast<float>(subpixels.size());
}

Color Pixel::GetVarianceColor() const
{
    Color mean = GetMeanColor();
    Color var = Color().Black();
    for (auto & subpix : subpixels)
    {
        var += subpix.color - mean;
    }
    return var / static_cast<float>(subpixels.size());
}