#pragma once
#ifndef PIXEL_H_
#define PIXEL_H_

#pragma warning (push, 1)
#include <scene.h>
#pragma warning (pop)
#include <vector>
#include <fstream>
#include <Serialization.h>

// stores all the information obtained after ray tracing sub-pixel samples
struct SubPixel
{
    float rx, ry;
    Color color;
public:
    void WriteToStream(std::ofstream & ofs) const
    {
        WritePrimitive(ofs, rx);
        WritePrimitive(ofs, ry);
        WritePrimitive(ofs, color.r);
        WritePrimitive(ofs, color.g);
        WritePrimitive(ofs, color.b);
    }
    void ReadFromStream(std::ifstream & ifs)
    {
        ReadPrimitive(ifs, rx);
        ReadPrimitive(ifs, ry);
        ReadPrimitive(ifs, color.r);
        ReadPrimitive(ifs, color.g);
        ReadPrimitive(ifs, color.b);
    }
};

struct Pixel
{
public:
    Pixel();
    void AddSubpixel();
    void AddSubpixelMax();
    Color GetMeanColor() const;
    Color GetVarianceColor() const;
    static const int basex = 2;
    static const int basey = 3;
    int x, y;
    std::vector<SubPixel> subpixels;
    Color color;
    void WriteToStream(std::ofstream & ofs) const
    {
        WritePrimitive(ofs, subpixels.size());
        for (auto & subpix : subpixels)
        {
            subpix.WriteToStream(ofs);
        }
    }
    void ReadFromStream(std::ifstream & ifs)
    {
        std::size_t size;
        ReadPrimitive(ifs, size);
        subpixels.resize(size);
        for (auto & subpix : subpixels)
        {
            subpix.ReadFromStream(ifs);
        }
    }
};

#endif