#pragma once
#ifndef RANDOM_H_
#define RANDOM_H_

#pragma warning (push, 1)
#include <scene.h>
#pragma warning (pop)
#include <random>
#include <common.h>
#include <thread>
#include <ctime>

#define thread_local __declspec( thread )

//inline auto Mersenne() -> std::mt19937_64 &
//{
//    static auto mersenne = std::mt19937_64(1234);
//    return mersenne;
//}
//
//inline auto Uniform() -> std::uniform_real_distribution<float> &
//{
//    static auto uniform = std::uniform_real_distribution<float>();
//    return uniform;
//}

// return a random float from 0 to 1
float RandFloat();
int RandInt(); // random integer from 0 to 2 billion

inline Point3 RandomPointInSphere(const Point3 & p, float r)
{
    float x = 0;
    float y = 0;
    float z = 0;
    do
    {
        x = (RandFloat() * 2 - 1) * r;
        y = (RandFloat() * 2 - 1) * r;
        z = (RandFloat() * 2 - 1) * r;
    } while (Square(x) + Square(y) + Square(z) > Square(r));
    return p + Point3(x, y, z);
}

inline Ray RandomRayOnSphere(const Point3 & p)
{
    Ray ray;
    ray.p = p;
    float r1 = RandFloat();
    float r2 = RandFloat();
    ray.dir.x = 2 * cos(TWO_PI * r1) * sqrt(r2 * (1 - r2));
    ray.dir.y = 2 * sin(TWO_PI * r1) * sqrt(r2 * (1 - r2));
    ray.dir.z = 1 - 2 * r2;
    return ray;
}

// u should be normalized
inline void BuildOrthonormalBasis(const Point3 & u, Point3 & v, Point3 & w)
{
    float x = std::abs(u.x);
    float y = std::abs(u.y);
    float z = std::abs(u.z);
    if (x <= y && x <= z)
    {
        v = Point3(0, -u.z, u.y);
    }
    else if (y <= x && y <= z)
    {
        v = Point3(-u.z, 0, u.x);
    }
    else
    {
        v = Point3(-u.y, u.x, 0);
    }
    v.Normalize();
    w = u.Cross(v);
}

inline void BuildOrthonormalBasisRandom(const Point3 & u, Point3 & v, Point3 & w)
{
    float x = RandFloat() * 2 - 1;
    float y = RandFloat() * 2 - 1;
    float z = RandFloat() * 2 - 1;
    v = Point3(x, y, z).GetNormalized();
    w = u.Cross(v);
    v = w.Cross(u);
}

inline Ray RandomRayInHemisphereCosine(const Point3 & p, const Point3 & n)
{
    Point3 v, w;
    BuildOrthonormalBasis(n, v, w);

    // generate random points on the disk
    float r = sqrt(RandFloat());
    float phi = RandFloat() * TWO_PI;
    // project these points to the hemisphere
    Point3 x = v * r * cos(phi);
    Point3 y = w * r * sin(phi);
    Point3 z = n * sqrt(1 - r * r);
    Point3 dir = (x + y + z).GetNormalized();

    return Ray(p, dir);
}

inline Ray HaltonRayInHemisphereCosine(const Point3 & p, const Point3 & n, int base1, int base2, int i)
{
    Point3 v, w;
    BuildOrthonormalBasisRandom(n, v, w);

    // generate random points on the disk
    float r = sqrt(Halton(i, base1));
    float phi = Halton(i, base2) * TWO_PI;
    // project these points to the hemisphere
    Point3 x = v * r * cos(phi);
    Point3 y = w * r * sin(phi);
    Point3 z = n * sqrt(1 - r * r);
    Point3 dir = (x + y + z).GetNormalized();

    return Ray(p, dir);
}

inline float PointDistToPlane(const Point3 & p, const Point3 & q, const Point3 & n)
{
    return fabs((p - q).Dot(n));
}

#endif