#pragma warning (push, 1)
#include <lights.h>
#pragma warning (pop)
#include <trace.h>
#include <common.h>
#include <functional>
#include <random>
#include <ctime>
#include <Sampling.h>
#include <iostream>

extern Node rootNode;

float GenLight::Shadow(Ray ray, float t_max)
{
    HitInfo hitInfo;
    hitInfo.z = t_max;
    hitInfo.shadow = true;
    bool isInShadow = TraceRay(ray, &rootNode, hitInfo, HIT_FRONT_AND_BACK);
    return isInShadow ? 0.0f : 1.0f;
}

// 8
//float poisson[] = { -0.4809184f, -0.462062f, 0.5653296f, -0.2167352f, -0.3168256f, 0.3430421f,
//                    -0.8897641f, -0.0664158f, 0.2529849f, -0.7175763f, 0.5744539f, 0.5629984f,
//                    0.1182156f, 0.9658515f, 0.001397021f, -0.1568614f };
// 16
//float poisson[] = { 0.6749641f, 0.4193309f, 0.5781788f, -0.03726418f,
//                    0.2242793f, 0.3036275f, 0.1791329f, 0.7716042f,
//                    0.07742687f, -0.06540706f, -0.1552138f, 0.4041522f,
//                    -0.5351333f, 0.2966464f, -0.219063f, 0.8822297f,
//                    -0.3049269f, -0.2883115f, -0.8215988f, 0.03141464f,
//                    0.3372906f, -0.3992189f, -0.5941688f, -0.5725858f,
//                    -0.08696189f, -0.7719472f, 0.7195633f, -0.6062809f,
//                    0.2992231f, -0.8508857f, 0.9929877f, 0.01225504f };

Color PointLight::Illuminate(const Point3 & p, const Point3 & N) const
{
    if (size > 0)
    {
        return IlluminateFromArea(p);
    }
    else
    {
        return IlluminateFromPoint(p);
    }
}

Color PointLight::IlluminateFromPoint(const Point3 & p) const
{
    // TODO: optimization for tracing shadow ray
    auto dir = position - p;
    float notShadow = Shadow(Ray(p, dir), 1);
    auto ret = intensity / (4 * PI * (0.0001f + dir.LengthSquared())) * notShadow;

    return ret;
}

Color PointLight::IlluminateFromArea(const Point3 & p) const
{
    // TODO: broken
    // find the vector from the point to the center of the light
    Point3 centerRay = (p - position).GetNormalized();

    // build a right-hand coordinate frame
    Point3 axis1 = centerRay.Cross(Point3(1, 0, 0)).GetNormalized();
    if (axis1.LengthSquared() < 1e-3f)
    {
        axis1 = centerRay.Cross(Point3(0, 1, 0)).GetNormalized();
    }
    Point3 axis2 = centerRay.Cross(axis1).GetNormalized();

    // select a random point on disk
    //float theta = rand() * 1.0f / RAND_MAX * TWO_PI;
    //float sinTheta = sin(theta);
    //float cosTheta = cos(theta);
    //float phi = TWO_PI * rand() * 1.0f / RAND_MAX + a; // 0 -> 4PI
    //float r = size * sqrt(rand() * 1.0f / RAND_MAX); // 0 -> radius
    //float x = r * sin(phi); float y = r * cos(phi);
    float x = 0, y = 0;
    do
    {
        x = (rand() * 1.0f / RAND_MAX * 2 - 1) * size;
        y = (rand() * 1.0f / RAND_MAX * 2 - 1) * size;
    } while (x * x + y * y > size * size);
    Point3 q = position + axis1 * x + axis2 * y;

    // trace this ray
    HitInfo hitInfo;
    hitInfo.z = 1;
    float illumVal = 0;
    bool isInShadow = TraceRay(Ray(p, q - p), &rootNode, hitInfo, HIT_FRONT_AND_BACK);
    if (!isInShadow) { ++illumVal; }
    return intensity * (illumVal);
}

void GenLight::SetViewportParam(int lightID, ColorA ambient, ColorA intensity, Point4 pos) const
{
    // empty
}

Ray PointLight::RandomPhoton() const
{
    return RandomRayOnSphere(position);
}

Color DirectAreaLight::Illuminate(const Point3 & p, const Point3 & N) const
{
    if (N.Dot(direction) > 0) return Color(0, 0, 0);
    /*float x, y, z;
    int totalNotShadow = 0;
    const int numSamples = 16;
    const float r = 3.0f / PointDistToPlane(p, center, direction);
    for (int i = 0; i < numSamples; ++i)
    {
    do
    {
    x = r * (RandFloat() * 2 - 1);
    y = r * (RandFloat() * 2 - 1);
    z = r * (RandFloat() * 2 - 1);
    } while (x * x + y * y + z * z > r * r);
    Point3 dir = -direction + Point3(x, y, z);
    totalNotShadow += Shadow(Ray(p, dir));
    }

    return intensity * totalNotShadow / (PI * radius * radius * numSamples);*/
    auto s = Shadow(Ray(p, -direction));
    return intensity * s / (PI * radius * radius);
}

Ray DirectAreaLight::RandomPhoton() const
{
    // generate a random point inside the disc
    Point3 u = direction;
    Point3 v, w;
    BuildOrthonormalBasis(u, v, w);
    float x, y;
    HitInfo hInfo;
    bool hit = false;
    do
    {
        x = radius * (RandFloat() * 2 - 1);
        y = radius * (RandFloat() * 2 - 1);
    } while (x * x + y * y > radius * radius);
    auto p = center + v * x + w * y;
    return Ray(p, direction);
}

Ray DirectAreaLight::RandomPhoton(const Point3 & p) const
{
    //float dist = (p - center).Dot(direction);
    Point3 q = p - direction;
    return Ray(q, direction);
}

Color AreaLight::Illuminate(const Point3 & p, const Point3 & N) const
{
    // generate random point inside the disc
    Point3 u = direction;
    Point3 v, w;
    BuildOrthonormalBasisRandom(u, v, w);
    float x, y;
    float totalNotShadow = 0;
    const int numSamples = 128;
    // NOTE: this is just an approximation, because we are not taking into account the BRDF
    int j = 0;
    for (int i = 0; i < numSamples; ++i)
    {
        do
        {
            /*x = radius * (RandFloat() * 2 - 1);
            y = radius * (RandFloat() * 2 - 1);*/
            x = radius * (Halton(j, 2) * 2 - 1);
            y = radius * (Halton(j, 3) * 2 - 1);
            ++j;
        } while (x * x + y * y > radius * radius);
        auto q = center + v * x + w * y;
        totalNotShadow += Shadow(Ray(p, q - p), 1);
    }
    return intensity * totalNotShadow / (2 * PI * PI * radius * radius * numSamples);
}

Ray AreaLight::RandomPhoton() const
{
    // generate a random point inside the disc
    Point3 u = direction;
    Point3 v, w;
    BuildOrthonormalBasis(u, v, w);
    float x, y;
    HitInfo hInfo;
    bool hit = false;
    do
    {
        x = radius * (RandFloat() * 2 - 1);
        y = radius * (RandFloat() * 2 - 1);
    } while (x * x + y * y > radius * radius);
    auto p = center + v * x + w * y;
    // generate a random direction (cosine-weighted)
    auto ray = RandomRayInHemisphereCosine(p, direction);
    return ray;
}